use anyhow::{anyhow, Context, Result};
use clap::Parser;
use std::{
    fs::File,
    io::{BufRead, BufReader, Write},
    net::{TcpListener, TcpStream},
    process::Command,
    sync::mpsc::channel,
    thread,
    time::Duration,
};

mod rmframes;
mod streamlsb;
use rmframes::RMFrames;
use streamlsb::StreamLSB;

#[derive(Parser, Debug)]
#[command(author, version)]
pub struct Opts {
    /// Listen for an (unsecure) TCP connection to send the data to.
    /// TODO: implement basic encryption
    #[arg(long, name = "port", short = 'l')]
    listen: usize,
}

// Parameters only for RM2, firmware version >= 3.7.0.1930

/// Height (in pixels) of the framebuffer.
const HEIGHT: usize = 1404;
/// Width (in pixels) of the framebuffer.
const WIDTH: usize = 1872;
/// How many bytes represent one pixel in the framebuffer.
const BYTES_PER_PIXEL: usize = 2;

fn main() -> Result<()> {
    let ref opts: Opts = Opts::parse();

    let (file, offset) = {
        let pid = xochitl_pid()?;
        let offset = rm2_fb_offset(pid)?;
        let mem = format!("/proc/{}/mem", pid);
        (mem, offset)
    };

    let src = RMFrames::init(&file, offset, WIDTH * HEIGHT * BYTES_PER_PIXEL)?;
    let src = StreamLSB::new(src);
    let mut src = src;

    let mut tgt: Box<dyn Write> = Box::new(listen_timeout(opts.listen, Duration::from_secs(60))?);

    std::io::copy(&mut src, &mut tgt).unwrap();

    return Ok(());
}

fn listen_timeout(port: usize, timeout: Duration) -> Result<TcpStream> {
    let listen_addr = format!("0.0.0.0:{}", port);
    let listen = TcpListener::bind(&listen_addr)?;
    eprintln!("[rM] listening for a TCP connection on {}", listen_addr);

    let (tx, rx) = channel();
    thread::spawn(move || {
        tx.send(listen.accept()).unwrap();
    });

    let (conn, conn_addr) = rx
        .recv_timeout(timeout)
        .context("Timeout while waiting for host to connect to reMarkable")??;
    eprintln!("[rM] connection received from {}", conn_addr);
    conn.set_write_timeout(Some(timeout))?;
    Ok(conn)
}

fn xochitl_pid() -> Result<usize> {
    let output = Command::new("/bin/pidof")
        .args(&["xochitl"])
        .output()
        .context("Failed to run `/bin/pidof xochitl`")?;
    if output.status.success() {
        let pid = &output.stdout;
        let pid_str = std::str::from_utf8(pid)?.trim();
        pid_str
            .parse()
            .with_context(|| format!("Failed to parse xochitl's pid: {}", pid_str))
    } else {
        Err(anyhow!(
            "Could not find pid of xochitl, is xochitl running?"
        ))
    }
}

fn rm2_fb_offset(pid: usize) -> Result<usize> {
    let file = File::open(format!("/proc/{}/maps", &pid))?;
    let line = BufReader::new(file)
        .lines()
        .skip_while(|line| matches!(line, Ok(l) if !l.ends_with("/dev/fb0")))
        .skip(1)
        .next()
        .with_context(|| format!("No line containing /dev/fb0 in /proc/{}/maps file", pid))?
        .with_context(|| format!("Error reading file /proc/{}/maps", pid))?;

    let addr = line
        .split("-")
        .next()
        .with_context(|| format!("Error parsing line in /proc/{}/maps", pid))?;

    let address = usize::from_str_radix(addr, 16).context("Error parsing framebuffer address")?;
    Ok(address + 8)
}
