use std::io::Read;

const BYTE_FACTOR: usize = 2;

/// Read every other byte from `R`.
///
/// If `R` is a stream of u16be, this should return the least-significant byte of each u16.
pub struct StreamLSB<R: Read> {
    /// The source to read from
    src: R,

    /// Number of filled bits in fragbuf
    arity: usize,

    /// Hold a partial u16, in case we read half of one
    fragbuf: [u8; BYTE_FACTOR],
}

impl<R: Read> StreamLSB<R> {
    pub fn new(src: R) -> Self {
        Self {
            src,
            arity: 0,
            fragbuf: [0; BYTE_FACTOR],
        }
    }
}

impl<R: Read> Read for StreamLSB<R> {
    fn read(&mut self, buf: &mut [u8]) -> std::io::Result<usize> {
        let n_raw = self.src.read(buf)?;

        let mut in_cursor = 0;
        let mut out_cursor = 0;

        if self.arity != 0 {
            // Number of bytes we need to read
            let need_read = BYTE_FACTOR - self.arity;
            let can_read = need_read.min(buf.len());
            self.fragbuf[self.arity..self.arity + can_read].copy_from_slice(&buf[0..can_read]);

            in_cursor += can_read;
            self.arity += can_read;
            self.arity %= BYTE_FACTOR;

            if self.arity != 0 {
                // We did not read a full set of bytes,
                // wait for more data.
                return Ok(0);
            } else {
                // Extract least-significant byte
                buf[0] = self.fragbuf[0];
                out_cursor += 1;
            }
        }

        for _ in in_cursor..(n_raw / BYTE_FACTOR) {
            // Extract least-significant byte
            buf[out_cursor] = buf[in_cursor];

            in_cursor += BYTE_FACTOR;
            out_cursor += 1;
        }

        self.arity = n_raw - in_cursor;
        self.fragbuf[0..self.arity].copy_from_slice(&buf[in_cursor..n_raw]);

        return Ok(out_cursor);
    }
}

// TODO: make this faster with SIMD

/*
SORKIN DEEPSEEK

On an ARMv7 processor with NEON SIMD support, you can significantly speed up the process of extracting and packing the most significant bits (MSBs) of every other byte from 16 consecutive bytes. NEON allows you to process multiple bytes in parallel using wide registers and specialized instructions.Here’s how you can achieve this using NEON intrinsics:### Steps:

    Load the 16 bytes into a NEON register.
    Extract the MSB of every byte using a right shift (`vshrq_u8`).
    Mask out every other byte using a bitwise AND (`vandq_u8`).
    Pack the MSBs into a single byte using NEON's horizontal reduction.

### Code Example:
```c
#include <arm_neon.h>
#include <stdint.h>
#include <stdio.h>uint8_t pack_msbs_neon(const uint8_t *data) {
    // Load 16 bytes into a NEON register
    uint8x16_t vec = vld1q_u8(data);    // Shift each byte right by 7 to isolate the MSB
    uint8x16_t msbs = vshrq_n_u8(vec, 7);    // Create a mask to select every other byte (0x80, 0x00, 0x80, 0x00, ...)
    uint8x16_t mask = {0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00,
                       0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00};    // Apply the mask to keep only every other MSB
    uint8x16_t masked_msbs = vandq_u8(msbs, mask);    // Horizontally add the bytes to pack the MSBs into a single byte
    uint8x8_t sum = vpadd_u8(vget_low_u8(masked_msbs), vget_high_u8(masked_msbs));
    sum = vpadd_u8(sum, sum);
    sum = vpadd_u8(sum, sum);    // Extract the final packed byte
    return vget_lane_u8(sum, 0);
}int main() {
    uint8_t data[16] = {
        0b10000001, 0b00000000, 0b11000011, 0b00000000,
        0b10101010, 0b00000000, 0b11110000, 0b00000000,
        0b10000001, 0b00000000, 0b11000011, 0b00000000,
        0b10101010, 0b00000000, 0b11110000, 0b00000000
    };    uint8_t packed = pack_msbs_neon(data);    printf("Packed MSBs: 0x%02X\n", packed);    return 0;
}
```### Explanation:

    **`vld1q_u8(data)`**: Loads 16 bytes from `data` into a NEON register (`uint8x16_t`).
    **`vshrq_n_u8(vec, 7)`**: Shifts each byte right by 7, isolating the MSB.
    **Masking**: A mask (`0x80` for every other byte) is applied using `vandq_u8` to keep only the MSBs of every other byte.
    **Horizontal Addition**: The `vpadd_u8` instruction is used to horizontally add pairs of bytes, effectively packing the MSBs into a single byte.
    **Final Extraction**: The result is extracted using `vget_lane_u8`.

### Performance:

    This approach processes 16 bytes in parallel using NEON, making it much faster than a scalar implementation.
    The use of horizontal addition (`vpadd_u8`) efficiently reduces the NEON register to a single byte.

### Example Output:
For the given `data` array, the output will be:
```
Packed MSBs: 0xAA
```This matches the expected result, where the MSBs of every other byte are packed into a single byte.### Compilation:
Make sure to enable NEON support when compiling:
```bash
gcc -o neon_example neon_example.c -mfpu=neon -march=armv7-a
```This code is optimized for ARMv7 with NEON and should provide a significant performance boost over scalar implementations.
*/
